/*
 * Process-related convenience functions.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>

#include <stdio.h>
#include <stdlib.h>

#include "process_creator_context_client.h"

#include "file.h"
#include "process.h"
#include "notifier.h"



/* Create a new process object. */

process_t *process_new()
{
  process_t *process = (process_t *) malloc(sizeof(process_t));

  if (process == NULL)
    return NULL;

  process_init(process);
  return process;
}

/* Discard the given process. */

void process_close(process_t *process)
{
  if (process == NULL)
    return;

  if (l4_is_valid_cap(process->ref))
    ipc_cap_free_um(process->ref);

  process_init(process);
}

/* Initialise the given process structure. */

void process_init(process_t *process)
{
  process->ref = L4_INVALID_CAP;
}

/* Start a process using the given arguments.
   NOTE: This does not yet obtain input/output pipes. */

long process_start(process_t *process, int argc, char *argv[])
{
  l4_cap_idx_t server = l4re_env_get_cap("pserver");

  if (l4_is_invalid_cap(server))
    return -L4_ENOMEM;

  /* Obtain a context for process creation. */

  file_t context;
  long err = file_context(&context, server);

  if (err)
    return err;

  offset_t pos = 0, written;

  for (int i = 0; i < argc; i++)
  {
    if (!file_string_set(&context, argv[i], pos, &written))
      return -L4_ENOMEM;

    pos += written + 1;
  }

  /* Obtain a client for the process creator context. */

  client_ProcessCreatorContext creator(context.ref);

  /* Start the process, obtaining a reference to it. */

  err = creator.start(argc, &process->ref);

  /* Initialise the notifiable section of the structure. */

  process->notifiable.notifications = 0;
  process->notifiable.base = (notifiable_base_t *) process;

  /* Close the context, although a separate mechanism could permit contexts to
     open several processes. */

  file_close(&context);
  return err;
}



/* NOTE: Much of the code below could be unified with the file-related
         notification code. */

/* Opaque notifier type for process_notifier_t. */

struct process_notifier
{
  ObjectNotifier *obj;
};

/* Conversion to the generic notification types. */

notifiable_t *process_notifiable(process_t *process)
{
  return &process->notifiable;
}

/* Return the notification flags for a process. */

notify_flags_t process_notifications(process_t *process)
{
  return process->notifiable.notifications;
}

/* Return the notification values for a process. */

notify_values_t process_notification_values(process_t *process)
{
  return process->notifiable.values;
}

/* Close a notifier object. */

void process_notify_close(process_notifier_t *notifier)
{
  delete notifier->obj;
  delete notifier;
}

/* Obtain a local notifier object. */

process_notifier_t *process_notify_local()
{
  process_notifier_t *notifier = new process_notifier_t;

  notifier->obj = notifier_get_local_notifier();
  return notifier;
}

/* Obtain the task-wide notifier object. */

process_notifier_t *process_notify_task()
{
  process_notifier_t *notifier = new process_notifier_t;

  notifier->obj = notifier_get_task_notifier();
  return notifier;
}

/* Subscribe to notification events on a process. */

long process_notify_subscribe(process_t *process, notify_flags_t flags, process_notifier_t *notifier)
{
  return notifier->obj->subscribe(process_notifiable(process), flags);
}

/* Unsubscribe from notification events on a process. */

long process_notify_unsubscribe(process_t *process, process_notifier_t *notifier)
{
  return notifier->obj->unsubscribe(process_notifiable(process));
}

/* Wait for a notification event on a process. */

long process_notify_wait_process(process_t *process, process_notifier_t *notifier)
{
  SpecificObjectNotifier *specific_notifier = dynamic_cast<SpecificObjectNotifier *>(notifier->obj);
  long err = specific_notifier->wait_object(process_notifiable(process));

  /* Unsubscribe if a termination notification has been received. */

  if (!err && (process->notifiable.notifications & NOTIFY_TASK_SIGNAL) &&
              (process->notifiable.values.sig == 0))
  {
    process_notify_unsubscribe(process, notifier);
    process_close(process);
  }

  return err;
}

/* Wait for notification events on processes. */

long process_notify_wait_processes(process_t **process, process_notifier_t *notifier)
{
  GeneralObjectNotifier *general_notifier = dynamic_cast<GeneralObjectNotifier *>(notifier->obj);
  notifiable_t *notifiable;
  long err = general_notifier->wait(&notifiable);

  *process = (process_t *) notifiable->base;

  /* Unsubscribe if a termination notification has been received. */

  if (!err && ((*process)->notifiable.notifications & NOTIFY_TASK_SIGNAL) &&
              ((*process)->notifiable.values.sig == 0))
  {
    process_notify_unsubscribe(*process, notifier);
    process_close(*process);
  }

  return err;
}

// vim: tabstop=2 expandtab shiftwidth=2
